from typing import Union
import sys
import os

from ..Log import Log
from .. import Func
from .WindowManager import WindowManager
from ..Graphic.Widgets import Widget
from ..Graphic.Windows import Window
from ..Classes import Event
from ..Config import Config

from ..Graphic.Windows import *
from ..Graphic.Widgets import *

class Parser:
    _file_path: Union[str, None] = None
    _file_update_time: Union[float, None] = None
    _builded_event: Event = Event()
    
    @classmethod
    def setFile(cls, path: str):
        if not os.path.exists(path):
            Log.writeWarning(f'File "{path}" not found', cls)
            return
        
        cls._file_path = path
        
        Log.writeOk(f'File setted', cls)
        
        cls.checkUpdate()
        
    
    @classmethod
    def checkUpdate(cls):
        if cls._file_path is None:
            return
        
        now_file_update_time = os.path.getmtime(cls._file_path)
        
        if now_file_update_time != cls._file_update_time:
            cls._file_update_time = now_file_update_time
            
            WindowManager.closeAll()
            Widget.removeAll()
            
            try:
                for window_data in Func.readJson(cls._file_path):
                    WindowManager.openNewWindow(
                        cls.buildWindow(window_data)
                    )
                    
                cls._builded_event.invoke()
                Log.writeOk('windows builded!', cls)
            except:
                WindowManager.closeAll()
                Widget.removeAll()
                Log.writeError()
    
    @classmethod
    def buildWindow(cls, window_data: dict[str, any]) -> Window:
        def parseWidgets(data: list[dict[str, any]]) -> list[Widget]:
            widgets: list[Widget] = []
            for widget_data in data:
                widget: type[Widget] = getattr(sys.modules['FloriaConsoleGUI.Graphic.Widgets'], widget_data['class'])
                widget_data.pop('class')
                
                for attr in widget_data:
                    if Config.PARSER_SKIP_UNKNOWED_ANNOTATIONS and attr not in widget.__init__.__annotations__:
                        Log.writeNotice(f'widget "{widget.__name__}" attribute "{attr}" skipped', cls)
                        continue
                    
                    attr_value = widget_data[attr]
                    match attr:
                        case 'widgets':
                            widget_data[attr] = parseWidgets(attr_value)
                    
                widgets.append(widget(**widget_data))
            return widgets
                
        window: type[Window] = getattr(sys.modules['FloriaConsoleGUI.Graphic.Windows'], window_data['class'])
        
        if 'widgets' in window_data:
            window_data['widgets'] = parseWidgets(window_data['widgets'])

        return window(**window_data)
    
    @classmethod
    def get_builded_event(cls) -> Event:
        return cls._builded_event


import os, sys

from ..Log import Log
from ..Threads import BaseThread
from ..Managers.WindowManager import WindowManager
from ..Graphic.Pixel import Pixel, Pixels
from ..Graphic.Widgets import Widget
from ..Graphic.Windows import Window
from ..Config import Config
from .. import Func

class GraphicThread(BaseThread):
    def __init__(self):
        super().__init__(1/Config.FPS)
        self._info = {}
    
    async def simulation(self):
        buffer = WindowManager.render()
        if buffer is None:
            return
        
        buffer_data = [pixel if pixel is not None else Pixel.empty for pixel in buffer.data]
        
        pixels: list[Pixel] = \
        [
            buffer_data[i].ANSII if i - i // buffer.width * buffer.width == 0 or not Pixel.compareColors(buffer_data[i-1], buffer_data[i]) else buffer_data[i].symbol 
            #buffer_data[i].symbol
            #buffer_data[i].ANSII
            for i in range(len(buffer_data))
        ]
        
        rendered_text = ''.join([
            ''.join(pixels[y*buffer.width : y*buffer.width+buffer.width]) + f'{Pixel.clearANSII}\n' for y in range(buffer.height)
        ])
        
        if Config.DEBUG_SHOW_DEBUG_DATA:
            if Func.every('update_info', 1, True):
                self._info = self.__class__._amt_sim.getAll()
                self.__class__._amt_sim.clearAll()
            
            Config.debug_data.update(self._info)
        
        sys.stdout.write(f'{'\n' * Config.CLEAR_LINES}{rendered_text}{'; '.join([f'{key}={value}' for key, value in Config.debug_data.items()]) if Config.DEBUG_SHOW_DEBUG_DATA else ''}')
    


from typing import Union, Iterable

from .Widget import Widget
from ..Pixel import Pixel, Pixels
from ...Classes import Vec2, Vec3, Vec4, Anchor
from ... import Func
from ... import Converter

class Label(Widget):
    def __init__(
        self, 
        text: str = 'label', 
        text_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        name: Union[str, None] = None,
        min_size: Union[Vec2[Union[int, None]], Iterable[Union[int, None]], None] = None,
        anchor: Union[Anchor, str] = Anchor.left,
        tab_length: int = 4,
        *args, **kwargs
    ):
        super().__init__(
            size=(0, 0), 
            padding=padding,
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            *args, **kwargs
        )
        
        self._text_pixel = Converter.toPixel(text_pixel)
        self._anchor = Converter.toAnchor(anchor)
        self._min_size = Converter.toVec2(min_size, Vec2(None, None), True)
        self._lines: tuple[str] = ()
        self._tab_length = tab_length
        
        self.text = text
    
    def refresh(self):
        super().refresh()
        
        text_pixel = Func.choisePixel(self.text_pixel, self.clear_pixel, Pixels.wt)
        for y in range(len(self._lines)):
            for x in range(len(self._lines[y])):
                self._buffer[
                    x + self.padding.left, 
                    y + self.padding.top
                ] = Pixel.changePixel(text_pixel, symbol=self._lines[y][x])
    
    def setText(self, value: str):
        text = str(value).replace('\t', ' '*self._tab_length)
        self._lines = tuple([
            Func.setTextAnchor(
                line, 
                self._anchor, 
                max(len(line), self._min_size.x),
                self._clear_pixel.symbol
            ) if self._min_size.x is not None else line for line in [
                line.replace('\n', '') for line in text.rsplit('\n')
            ]
        ])
        
        self.size = Vec2(
            max(map(len, self._lines)), 
            len(self._lines)
        )
        
        self.setFlagRefresh()
        
    def getText(self) -> str:
        return '\n'.join(self._lines)
    @property
    def text(self) -> str:
        return self.getText()
    @text.setter
    def text(self, value: str):
        self.setText(value)
    
    @property
    def text_pixel(self) -> Union[Pixel, None]:
        return self._text_pixel
    @text_pixel.setter
    def text_pixel(self, value: Union[Pixel, None]):
        self._text_pixel= value




from typing import Union, Iterable, Unpack

from ...Classes import Vec2, Vec3, Event, Buffer
from .Widget import Widget
from ..Pixel import Pixel
from ... import Converter

class InteractiveWidget(Widget):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None, 
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        name: Union[str, None] = None,
        tabindex: Union[int, None] = None,
        select_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str, None] = None,
        *args, **kwargs):
        super().__init__(size, offset_pos, clear_pixel, name, *args, **kwargs)
        
        self._change_tabindex_event: Event = Event()
        self._tabindex: Union[int, None] = None
        self.tabindex = tabindex
        
        self._select_pixel = Converter.toPixel(select_pixel)
        self._selected: bool = False
    
    def refresh(self):
        if self._selected:
            self._buffer = Buffer(*self._size, self._select_pixel)
            print(f'interactive widtget refreshed {self}')
        else:
            super().refresh()
    
    @property
    def tabindex(self) -> Union[int, None]:
        return self._tabindex
    @tabindex.setter
    def tabindex(self, value: Union[int, None]):
        if value is not None and value < 0:
            raise ValueError('Index cannot be less than 0')
        self._tabindex = value
        self._change_tabindex_event.invoke()
    
    @property
    def change_tab_index_event(self) -> Event:
        return self._change_tabindex_event

    @property
    def selected(self) -> bool:
        return self._selected
    @selected.setter
    def selected(self, value: bool):
        if self._selected is not value:
            self._selected = value
            self.setFlagRefresh()        
    
    def __str__(self, **kwargs):
        kwargs.update({
            "tabindex": self._tabindex,
        })
        return super().__str__(**kwargs)
        
        


from typing import Union, Iterable
import os
import PIL
import PIL.GifImagePlugin
import PIL.Image

from .. import Func
from ..Graphic import Animation, Pixel
from ..Classes import Buffer


class AnimationManager:
    _animations: dict[str, Animation] = {}

    @classmethod
    def get(cls, name: str) -> Animation:
        if name not in cls._animations:
            raise ValueError(f'Animation "{name}" not found')
        return cls._animations[name].copy()
    
    @classmethod
    def register(cls, name: str, animation: Animation):
        if name in cls._animations:
            raise ValueError('//')
        cls._animations[name] = animation
    
    @classmethod
    def load(cls, path: str):
        '''
            Path to a json file with data like:\n
            [
                {
                    name: `str` - animation name,
                    path: `str` - path to img/jpg/jpeg/gif,
                    delay: `float` default=get from source file - delay between frames,
                    loop: `bool` default=True - animation is looped
                }, \n
                ...
            ]
        '''
        dir_path = os.path.dirname(path)
        
        animations_data: list[dict[str, any]] = Func.readJson(path)
        for data in animations_data:
            image = PIL.Image.open(f'./{dir_path}/{data['path']}', formats=['png', 'jpeg', 'gif'])
            
            cls.register(
                data['name'],
                Animation(
                    delay=data.get('delay', cls.getDelay(image)),
                    loop=data.get('loop', True),
                    frames=cls.getFrames(image)
                )
            )
    
    @classmethod
    def getDelay(cls, image: Union[PIL.Image.Image, PIL.GifImagePlugin.GifImageFile]) -> float:
        if isinstance(image, PIL.GifImagePlugin.GifImageFile):
            return image.info.get('duration') / 1000
        return 0
    
    @classmethod
    def getFrames(cls, image: Union[PIL.Image.Image, PIL.GifImagePlugin.GifImageFile]) -> Iterable[Buffer[Pixel]]:
        def convertImageToBuffer(image: Union[PIL.Image.Image, PIL.GifImagePlugin.GifImageFile]) -> Buffer[Pixel]:
            return Buffer(
                image.width,
                image.height,
                None,
                [
                    Pixel(back_color=pixel_data[:3]) if pixel_data[3] > 125 else None for pixel_data in image.convert('RGBA').getdata()
                ]
            )
        
        if isinstance(image, PIL.GifImagePlugin.GifImageFile):
            frames = []
            for i in range(image.n_frames):
                image.seek(i)
                frames.append(
                    convertImageToBuffer(image)
                )
            return frames
        
        else:
            return [
                Buffer(
                    image.width,
                    image.height,
                    None,
                    convertImageToBuffer(image)
                )
            ]
                
    


from typing import Union, Callable, Generic, TypeVar, Iterable
import math
from copy import deepcopy

from .Vec import Vec2

_T = TypeVar('_T')

class Buffer(Generic[_T]):
    empty = None
    
    def __init__(self, width: int, height: int, defualt_value: _T = None, data: Union[Iterable[_T], None] = None):
        if width < 0 or height < 0:
            raise ValueError(f'Width or height cannot be less than 0')
        
        self._size = Vec2(width, height)
        self._size.change_event.add(self._raise_resize)
        
        self._defualt_value = defualt_value
        
        if data is not None:
            if len(data) != width*height:
                raise ValueError(f'The parameters for width({width}) and height({height}) do not match the len of data({len(data)})')
            self._data: list[_T] = list(data)  
        else:
            self.fill()


    def paste(self, offset_x: int, offset_y: int, buffer: Union['Buffer', None], func: Callable[[_T, _T], _T] = lambda old, new: new if new is not None else old):
        '''
            func: function(self_value, other_value) -> new_value
        '''
        if buffer is None or buffer.size.x == 0 or buffer.size.y == 0:
            return
        
        for y in range(buffer.height):
            for x in range(buffer.width):
                pos = Vec2(
                    offset_x + x, 
                    offset_y + y
                )
                if 0 <= pos.x < self.width and 0 <= pos.y < self.height:
                    self[*pos] = func(self.get(*pos), buffer[x, y])
        

    def fill(self, value: _T = None):
        '''
            Fill default_value if value is None
        '''
        self._data = [value if value is not None else self._defualt_value] * (self.width * self.height)
    
    def set(self, x: int, y: int, value: _T):
        self._data[y * self.width + x] = value
    def get(self, x: int, y: int) -> _T:
        return self._data[y * self.width + x]
    
    def convert(self, func: Callable[[_T], any]) -> 'Buffer':
        '''
            Create and convert a buffer\n
            `Don't modify` this buffer, just `create a new one`
        '''
        
        return Buffer(*self.size, self._defualt_value, (func(item) for item in self._data))        
    
    def resize(self, width: int, height: int) -> 'Buffer':
        if self.size == (width, height):
            return self.copy()
        
        buffer = Buffer(width, height, self._defualt_value)
        
        xk = self.width / width 
        yk = self.height / height
        
        for y in range(buffer.height):
            for x in range(buffer.width):
                buffer[x, y] = self[
                    min(max(math.floor(x*xk), 0), self.width-1), 
                    min(max(math.floor(y*yk), 0), self.height-1)
                ]
        
        return buffer
    
    @property
    def size(self) -> Vec2[int]:
        return self._size
    @property
    def width(self) -> int:
        return self.size.width
    @property
    def height(self) -> int:
        return self.size.height
    @property
    def data(self) -> tuple[_T]:
        return tuple(self._data)
    
    def __len__(self) -> int:
        return len(self._data)
    
    def __iter__(self):
        yield from self._data
                 
    def __getitem__(self, pos: tuple[int, int]) -> _T:
        return self.get(*pos)
    def __setitem__(self, pos: tuple[int, int], value: _T):
        self.set(*pos, value)

    @staticmethod
    def _raise_resize():
        raise RuntimeError('Size cannot be changed')
    
    def copy(self) -> 'Buffer':
        return deepcopy(self)
    
Buffer.empty = Buffer(0, 0, None)


import asyncio
import os
import sys
from typing import Union
import importlib
import time



from .Config import Config
from .Threads import BaseThread, GraphicThread, SimulationThread, InputThread
from .Classes.Event import Event
from .Log import Log
from .Managers import KeyboardManager as KeyM, posix_key, WindowManager
from . import Func

class Core:   
    init_event: Event = Event()
    term_event: Event = Event()
    
    init_all_event = BaseThread.init_all_event
    
    GraphicThread = GraphicThread()
    SimulationThread = SimulationThread()
    InputThread = InputThread()

    _inited = False
    _tasks: list[asyncio.Task] = []
    
    @classmethod
    def start(cls):
        '''
            run async app
        '''
        try:
            if cls._inited is False:
                raise RuntimeError('Core was not initialized')
            
            for thread in BaseThread._threads.values():
                cls._tasks.append(Config.ASYNC_EVENT_LOOP.create_task(thread.run(), name=thread.name))
            
            Config.ASYNC_EVENT_LOOP.run_until_complete(asyncio.wait(cls._tasks))
        
        except KeyboardInterrupt:
            Log.writeWarning('Emergency termination', cls)
            
        except:
            Log.writeError(cls)
            input('Press to continue...')
                
        finally:
            for task in cls._tasks:
                task.cancel()
            Config.ASYNC_EVENT_LOOP.run_until_complete(asyncio.wait(cls._tasks))
            Config.ASYNC_EVENT_LOOP.stop()
            
    @classmethod
    def init(cls):
        if Config.CORE_MODIFY_WIN_REGEDIT and sys.platform == 'win32':
            import winreg
            
            try:
                reg_value = winreg.QueryValueEx(
                    winreg.OpenKey(
                        winreg.HKEY_CURRENT_USER, 
                        'Console', 
                        access=winreg.KEY_READ
                    ), 
                    'VirtualTerminalLevel'
                )
            except:
                reg_value = None
            
            if reg_value is None or reg_value[0] != 1:
                winreg.SetValueEx(
                    winreg.OpenKey(
                        winreg.HKEY_CURRENT_USER, 
                        'Console', 
                        access=winreg.KEY_SET_VALUE
                    ), 
                    'VirtualTerminalLevel',
                    0,
                    winreg.REG_DWORD,
                    1
                )
                
                input('Regedit has been modified, please restart the application\nPress to close...')
                exit()
                
        KeyM.registerEvent('_close', posix_key.CTRL_C)
        KeyM.bindEvent('_close', '\x00k') # alt+f4
        KeyM.bind('_close', BaseThread.stopAll)
        
        cls.init_event.invoke()
        
        cls._inited = True
        Log.writeOk('Initialized', cls)
    
    @classmethod
    def term(cls):
        if cls._inited is False:
            raise RuntimeError('Core was not initialized')
        
        cls.init_event.invoke()
        Log.writeOk('Terminated', cls)
    
    _dynamic_modules: dict[str, dict[str, any]] = {}
    @classmethod
    def addDynamicModule(cls, path: str, name: str):
        if path in cls._dynamic_modules:
            raise ValueError(f'Module "{path}" already exists')
        
        if not os.path.exists(path):
            raise ValueError(f'File "{path}" not exists')
        
        if Config.CORE_WRITE_WARNING_DYNAMIC_MODULE and len(cls._dynamic_modules) == 0:
            Log.writeWarning(
                '\nThis tool is unstable due to its features and may lead to errors\nIt is strongly recommended to only change variables inside the module\nNo complex logic', cls
            )
            time.sleep(0.5)
            
        
        cls._dynamic_modules[path] = {
            'mtime': os.path.getmtime(path),
            'name': name,
            'module': importlib.import_module(name)
        }
        Log.writeOk(f'module "{path}" added', cls)
        
    @classmethod
    def popDynamicModule(cls, path: str):
        if cls._dynamic_modules.pop(path, None) is not None:
            Log.writeOk(f'module "{path}" removed', cls)
    
    @classmethod
    def checkDynamicModules(cls):
        for path, data in cls._dynamic_modules.items():
            if not os.path.exists(path):
                continue
            
            os.path.dirname
            
            last_mtime = os.path.getmtime(path)
            if data['mtime'] == last_mtime:
                continue
            
            data['module'] = importlib.reload(data['module'])
            data['mtime'] = last_mtime
            
            Log.writeOk(f'module "{path}" updated')
            
    

             


from typing import Union, Iterable

from ..Classes import Vec2, Vec3, Vec4, Event, Buffer
from .Pixel import Pixel, Pixels
from .. import Converter


class BaseGraphicObject:
    def __init__(
        self,
        size: Union[Vec2[int], Iterable[int]] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        always_refresh: bool = False,
        always_parent_refresh: bool = False,
        *args, **kwargs
        ):
        
        ''' events '''
        self.__resize_event = Event(
            self.setFlagRefresh
        )
        self.__change_clear_pixel_event = Event(
            self.setFlagRefresh
        )
        self.__set_refresh_event = Event()
        
        ''' size and pos '''
        self._offset_pos = Converter.toVec3(offset_pos)
        self._size = Converter.toVec2(size)
        self._padding: Vec4[int] = Converter.toVec4(padding)
        
        ''' buffers '''
        self._buffer: Buffer[Pixel] = Buffer.empty
        
        ''' pixels '''
        self._clear_pixel = Converter.toPixel(clear_pixel)
        
        ''' flags '''
        self._flag_refresh = True
        
        ''' other '''
        self._name = name
        self._always_refresh = always_refresh
        self._always_parent_refresh = always_parent_refresh


    def refresh(self):
        self._buffer = Buffer(
            self.width + self.padding.left + self.padding.right,
            self.height + self.padding.top + self.padding.bottom,
            self.clear_pixel
        )
        
        self._flag_refresh = False
    
    def render(self) -> Buffer[Pixel]:
        if self._flag_refresh or self._always_refresh or self._always_parent_refresh:
            self.refresh()
        return self._buffer
    
    def setFlagRefresh(self):
        self._flag_refresh = True
        self.set_refresh_event.invoke()
    
    
    def setOffsetPos(self, value: Vec3[int]):
        self._offset_pos = value
    def getOffsetPos(self) -> Vec3[int]:
        return self._offset_pos
    @property
    def offset_pos(self) -> Vec3[int]:
        return self.getOffsetPos()
    @offset_pos.setter
    def offset_pos(self, value: Vec3[int]):
        self.setOffsetPos(value)
    @property
    def offset_x(self) -> int:
        return self.offset_pos.x
    @offset_x.setter
    def offset_x(self, value: int):
        self.offset_pos.x = value
    @property
    def offset_y(self) -> int:
        return self.offset_pos.y
    @offset_y.setter
    def offset_y(self, value: int):
        self.offset_pos.y = value
    @property
    def offset_z(self) -> int:
        return self.offset_pos.z
    @offset_z.setter
    def offset_z(self, value: int):
        self.offset_pos.z = value
    
    def setSize(self, value: Vec2[int]):
        self._size = value
        self.resize_event.invoke()
        value.change_event.add(
            self.resize_event.invoke
        )
    def getSize(self) -> Vec2[int]:
        return self._size
    @property
    def size(self) -> Vec2[int]:
        return self.getSize()
    @size.setter
    def size(self, value: Vec2[int]):
        self.setSize(value)
    @property
    def width(self) -> int:
        return self.size.width
    @width.setter
    def width(self, value: int):
        self.size.width = value
    @property
    def height(self) -> int:
        return self.size.height
    @height.setter
    def height(self, value: int):
        self.size.height = value
    
    @property
    def name(self) -> Union[str, None]:
        return self._name
    
    @property
    def clear_pixel(self) -> Union[Pixel, None]:
        return self._clear_pixel
    @clear_pixel.setter
    def clear_pixel(self, value: Union[Pixel, None]):
        self._clear_pixel = value
        self.change_clear_pixel_event.invoke()
    
    @property
    def resize_event(self) -> Event:
        return self.__resize_event
    @property
    def change_clear_pixel_event(self) -> Event:
        return self.__change_clear_pixel_event
    @property
    def set_refresh_event(self) -> Event:
        return self.__set_refresh_event
    
    def setPadding(self, value: Vec4[int]):
        self._padding = value
        self.__resize_event.invoke()
        value.change_event.add(
            self.__resize_event.invoke
        )
    def getPadding(self) -> Vec4[int]:
        return self._padding
    @property
    def padding(self) -> Vec4[int]:
        '''
            `up`: 0 | x\n
            `bottom` 1 | y\n
            `left` 2 | z\n
            `right` 3 | w
        '''
        return self.getPadding()
    @padding.setter
    def padding(self, value: Vec4[int]):
        self.setPadding(value)


class BaseGraphicContainerObject(BaseGraphicObject):
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None,
        padding: Union[Vec4[int], Iterable[int]] = None,
        auto_size: bool = False,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        objects: Union[Iterable[BaseGraphicObject], BaseGraphicObject] = [], 
        *args, **kwargs
    ):
        super().__init__(
            size=size,
            padding=padding,
            offset_pos=offset_pos, 
            clear_pixel=clear_pixel, 
            name=name, 
            *args, **kwargs
        )
        
        ''' size and pos '''
        self._auto_size = auto_size
        
        ''' events '''
        self.__add_object_event = Event(
            self.setFlagRefresh
        )
        
        ''' objects '''
        self._objects: list['BaseGraphicObject'] = []
        for object in Converter.toListObjects(objects):
            self.addObject(object)
        
        ''' buffers '''
        self._objects_buffer: Buffer[Pixel] = Buffer.empty
    
    def refresh(self):
        objects = [
            (
                object.offset_x,
                object.offset_y,
                object.render()
            )
            for object in self._objects
        ]
        
        super().refresh()
        
        self._objects_buffer = Buffer(
            max(self.width - sum(self.padding[2, 3]), 0),
            max(self.height - sum(self.padding[0, 1]), 0),
            self.clear_pixel
        )
        
        for object_data in objects:
            self._objects_buffer.paste(*object_data)
            
        self._buffer.paste(
            *self.padding[2, 0],
            self._objects_buffer
        )
    
        
    def addObject(self, object: BaseGraphicObject):
        self._objects.append(
            object
        )
        object.set_refresh_event.add(self.setFlagRefresh)
        self.add_object_event.invoke()
        if object._always_refresh or object._always_parent_refresh:
            self._always_refresh = True
    
    def getSize(self):
        return super().getSize()
    
    @property
    def add_object_event(self) -> Event:
        return self.__add_object_event
    
    def __iter__(self):
        yield from self._objects
    
    def __str__(self, **kwargs):
        kwargs.update({
            "name": self._name,
            "size": self._size,
            "offset_pos": self._offset_pos
        })
        return f'{self.__class__.__name__}({' '.join([f'{key}:{value}' for key, value in kwargs.items()])})'



from typing import Union, Iterable
import numpy as np 

from ..BaseGraphicObject import BaseGraphicObject, BaseGraphicContainerObject
from ..Pixel import Pixel, Pixels
from ..Drawer import Drawer
from ..Widgets import Widget
from ...Classes import Buffer, Event, Vec2, Vec3, Vec4

from ... import Converter
from ... import Func

class Window(BaseGraphicContainerObject):    
    def __init__(
        self, 
        size: Union[Vec2[int], Iterable[int]] = None,
        auto_size: bool = False,
        offset_pos: Union[Vec3[int], Iterable[int]] = None, 
        clear_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        name: Union[str, None] = None,
        widgets: Union[Iterable[Widget], Widget] = [], 
        frame: bool = False,
        frame_pixel: Union[Pixel, tuple[Union[Vec3[int], Iterable[int]], Union[Vec3[int], Iterable[int]], str], str] = None,
        *args, **kwargs
    ):
        super().__init__(
            size=size, 
            auto_size=auto_size, 
            offset_pos=offset_pos, 
            clear_pixel=Func.choisePixel(clear_pixel, Pixel.empty), 
            name=name, 
            objects=widgets, 
        *args, **kwargs)
        
        ''' events '''
        self._open_event = Event()
        self._close_event = Event()
        self._change_frame_pixel_event = Event()
        
        ''' pixels ''' 
        self._frame_pixel = Converter.toPixel(frame_pixel)
        
        ''' other '''
        self._frame = frame
    
    def refresh(self):
        super().refresh()
        
        if self.frame:
            frame_pixel: Pixel = Func.choisePixel(
                self.frame_pixel, 
                self.clear_pixel
            )
            
            self._buffer.paste(
                0, 0,
                Drawer.frame(
                    *self.size,
                    frame_pixel.front_color, 
                    frame_pixel.back_color
                )
            )
        
    def getPadding(self):
        return super().getPadding() + (
            Vec4(1, 1, 1, 1) if self.frame else Vec4(0, 0, 0, 0)
        )
    
    @property
    def open_event(self) -> Event:
        return self._open_event
    @property
    def close_event(self) -> Event:
        return self._close_event
    
    def setFrame(self, value: bool):
        self._frame = value
        self.setFlagRefresh()
    @property
    def frame(self) -> bool:
        return self._frame
    @frame.setter
    def frame(self, value: bool):
        self.setFrame(value)
    
    @property
    def frame_pixel(self) -> Union[Pixel, None]:
        return self._frame_pixel
    @frame_pixel.setter
    def frame_pixel(self, value):
        self._frame_pixel = value
        self.setFlagRefresh()
        self.change_frame_pixel_event.invoke()

    @property
    def change_frame_pixel_event(self) -> Event:
        return self._change_frame_pixel_event



from .Widget import Widget
from .Label import Label
from .Frame import Frame
from .Container import Container
from .ProgressBar import ProgressBar
from .Scroll import Scroll
from .InteractWidget import InteractiveWidget
from .Media import Media